/*
    Copyright (c) 2002 Florian Wesch <fw@dividuum.de>. All Rights Reserved.     
    
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.
    
    You should have received a copy of the GNU General Public License along
    with this program; if not, write to the Free Software Foundation, Inc.,
    59 Temple Place - Suite 330, Boston, MA  02111-1307, USA
*/    


#ifndef BRAINFUCKED_ASSEMBLER_H
#define BRAINFUCKED_ASSEMBLER_H

#include <stdlib.h>
#include <stdio.h>
#include "brainfuck.h"

typedef enum {  BFInst,         // Plain Instruction
                BFMove,         // Move absolute
                BFMoveRel,      // Move relative
                BFRecalibrateMP,// Sets MP to fixed value
                BFSetTo,        // Set Cell to value
                BFLoopS,        // Begin of loop
                BFLoopE,        // End of loop
                BFDebug         // Nonstandard BF debug
} bfEnum;

typedef struct BFInstructionStruct {
    bfEnum type;
    union {
        BFINST  bfinst;
        BFADDR  moveto;
        BFVALUE setto;
        BFADDR  newmp;
        int     moverelative;
        struct BFInstructionStruct *looppeer;
    } val;
    struct BFInstructionStruct *next;
    struct BFInstructionStruct *prev;
} BFInstruction;

typedef struct BFAssemblerContextStruct {
    int looplevel;
    BFInstruction *loopstack[MAXLOOPLEVEL];
    int instructions;
    int metainstructions;
    BFADDR freefortemp;
    BFInstruction *first;
    BFInstruction *cur;
} BFAssemblerContext;


// Funktionen

BFAssemblerContext *BF_AsmInit(BFADDR freefortemp);

void    BF_AllocNewInst(BFAssemblerContext *ctx);
void    BF_RemoveInst(BFAssemblerContext *ctx, BFInstruction *inst);

void    BF_AssemblerDump(BFAssemblerContext *ctx);
void    BF_Assemble(BFAssemblerContext *ctx, FILE *out);

BFADDR  BF_Newvar(BFAssemblerContext *ctx);
void    BF_Removevar(BFAssemblerContext *ctx, BFADDR mem_to_free);

void    BF_StartLoop(BFAssemblerContext *ctx);
void    BF_EndLoop(BFAssemblerContext *ctx);


#define BFADDR_OFFSET(addr, offset) \
    ((addr)+(offset))

// Makros zum anfgen einer neuen Anweisung
#define BFINST(inst) {              \
    BF_AllocNewInst(ctx);           \
    ctx->cur->type = BFInst;        \
    ctx->cur->val.bfinst = inst;    \
}

// Makro zum anfgen eines neuen moves von der 
// bisherigen position zur gegebenen
#define BFMOVE(dest) {              \
    BF_AllocNewInst(ctx);           \
    ctx->cur->type = BFMove;        \
    ctx->cur->val.moveto = dest;    \
}

// Makro zum anfgen einer Set Anweisung
#define BFSETTO(value) {            \
    BF_AllocNewInst(ctx);           \
    ctx->cur->type = BFSetTo;       \
    ctx->cur->val.setto = value;    \
}
    
// Makro zum starten eines neuen Loops 
#define BFLOOPS {                   \
    BF_AllocNewInst(ctx);           \
    ctx->cur->type = BFLoopS;       \
    BF_StartLoop(ctx);              \
}

// Makro zum beenden eines Loops 
#define BFLOOPE {                   \
    BF_AllocNewInst(ctx);           \
    ctx->cur->type = BFLoopE;       \
    BF_EndLoop(ctx);                \
}

// Makro fr relative Bewegungen 
#define BFMOVEREL(i) {              \
    BF_AllocNewInst(ctx);           \
    ctx->cur->type = BFMoveRel;     \
    ctx->cur->val.moverelative = i; \
}

// Makro um MP auf neuen festen Wert zu setzen
// Sinnvoll nach wilden MOVEREL Orgien
#define BFRECALIBRATEMP(i) {              \
    BF_AllocNewInst(ctx);           \
    ctx->cur->type = BFRecalibrateMP;\
    ctx->cur->val.newmp = i;        \
}

// Makro fr relative Bewegungen 
#define BFDEBUG {                   \
    BF_AllocNewInst(ctx);           \
    ctx->cur->type = BFDebug;       \
}

// Makro zum reservieren und freigeben von Speicherstellen
// fr temporre Nutzung
#define BFNEWVAR         BF_Newvar(ctx)
#define BFREMOVEVAR(mem) BF_Removevar(ctx, (mem))


// Makros fr die "high"level Funktionsdefinition und Funktionsausruf 

#define BFINST0(name) void BF_##name(BFAssemblerContext *ctx)
#define BFCALL0(name)                 \
    /* printf("call %s()\n", #name); */ \
    BF_##name(ctx)
    
#define BFINST1(name, arg1) void BF_##name(BFAssemblerContext *ctx, arg1)
#define BFCALL1(name, arg1)                 \
    /* printf("call %s(%d)\n", #name, (arg1)); */ \
    BF_##name(ctx, (arg1))
    
#define BFINST2(name, arg1, arg2) void BF_##name(BFAssemblerContext *ctx, arg1, arg2)
#define BFCALL2(name, arg1, arg2)                       \
    /* printf("call %s(%d, %d)\n", #name, (arg1), (arg2)); */ \
    BF_##name(ctx, (arg1), (arg2))
    
#define BFINST3(name, arg1, arg2, arg3) void BF_##name(BFAssemblerContext *ctx, arg1, arg2, arg3)
#define BFCALL3(name, arg1, arg2, arg3)                             \
    /*printf("call %s(%d, %d, %d)\n", #name, (arg1), (arg2), (arg3)); */ \
    BF_##name(ctx, (arg1), (arg2), (arg3))


// IF THEN Konstrukt

/*
    if(mem, <expr>)
        newvar(temp) copy(mem, temp) to(temp) [ <expr>  zero(temp) ] removevar(temp)
=>
    if_start(mem)
        newvar(temp) copy(mem, temp) to(temp) [ 
    if_end()
        zero(temp) ] removevar(temp)
*/
#define if_begin(mem)                           \
    { /* FIXME: macht if_loop_temp probleme? */ \
    BFADDR if_loop_temp = BFNEWVAR;             \
        BFCALL2(copy, mem, if_loop_temp);       \
        BFCALL1(to, if_loop_temp);              \
        BFINST(BF_OPLOOPS);             

#define if_end                                  \
            BFCALL1(zero, if_loop_temp);        \
        BFINST(BF_OPLOOPE);                     \
    BFREMOVEVAR(if_loop_temp);                  \
    }


// IF THEN ELSE Konstrukt

/*
    ifelse(mem, <ifexpr>, <elseexpr>)
        if(mem, <ifexpr>) newvar(temp) copy(mem, temp) not(temp) if(temp, <elseexpr>) removevar(temp)
=>
    ifelse_begin(mem) 
        if_begin(mem) 

    <ifexpr> 

    ifelse_else(mem)
        if_end
        newvar(ifelse_temp) 
        copy(mem, ifelse_temp) 
        not(ifelse_temp) 
        if_begin(ifelse_temp)
        
    <elseexpr>
    
    ifelse_end
        if_end
        removevar(ifelse_temp)
*/
#define ifelse_begin(mem)                       \
    {                                           \
    BFADDR ifelse_temp=BFNEWVAR;                \
        if_begin(mem);

#define ifelse_else(mem)                        \
        if_end;                                 \
        BFCALL2(copy, mem, ifelse_temp);        \
        BFCALL1(not, ifelse_temp);              \
        if_begin(ifelse_temp);                       

#define ifelse_end                              \
        if_end;                                 \
    BFREMOVEVAR(ifelse_temp);                   \
    }

// WHILE Konstrukt

/*
    while_begin(mem) 
        to(mem) [

    <whilebody> 

    while_end(mem)
        to(mem) ]
*/
#define while_begin(mem)                        \
    BFCALL1(to, mem);                           \
    BFLOOPS; 

#define while_end(mem)                          \
    BFCALL1(to, mem);                           \
    BFLOOPE;

// Brainfuck "High"level Befehle

BFINST1(to,                         BFADDR dest);    
BFINST1(zero,                       BFADDR dest); 
BFINST1(print,                      BFADDR mem); 
BFINST1(input,                      BFADDR dest);
BFINST1(inc,                        BFADDR dest);
BFINST1(dec,                        BFADDR dest);
BFINST1(left,                       int constvalue);
BFINST1(right,                      int constvalue);
BFINST2(set_const,                  BFADDR dest,    BFVALUE constvalue);

BFINST2(add_destroy,                BFADDR dest,    BFADDR value);
BFINST2(substract_destroy,          BFADDR dest,    BFADDR value);
BFINST2(move_destroy,               BFADDR source,  BFADDR dest);
BFINST3(move2_destroy,              BFADDR source,  BFADDR dest1,           BFADDR dest2);
BFINST2(copy,                       BFADDR source,  BFADDR dest);
BFINST2(mul,                        BFADDR memory,  BFADDR value);
BFINST2(mul_const,                  BFADDR memory,  BFVALUE constvalue);
BFINST1(not,                        BFADDR memory);
BFINST3(and,                        BFADDR source1, BFADDR source2,         BFADDR dest);
BFINST3(or,                         BFADDR source1, BFADDR source2,         BFADDR dest);
BFINST2(substractToNull_destroy,    BFADDR source1, BFADDR source2);
BFINST3(notEqual_destroy,           BFADDR source1, BFADDR source2,         BFADDR dest);
BFINST3(Equal_destroy,              BFADDR source1, BFADDR source2,         BFADDR dest);
BFINST3(Greater_destroy,            BFADDR source1, BFADDR source2,         BFADDR dest);
BFINST3(Less_destroy,               BFADDR source1, BFADDR source2,         BFADDR dest);
BFINST3(GreaterOrEqual_destroy,     BFADDR source1, BFADDR source2,         BFADDR dest);
BFINST3(LessOrEqual_destroy,        BFADDR source1, BFADDR source2,         BFADDR dest);

BFINST1(ArrayInit,                  BFADDR arraybase);
BFINST3(ArraySet,                   BFADDR arraybase,   BFADDR index,       BFADDR source);
BFINST3(ArrayGet,                   BFADDR arraybase,   BFADDR index,       BFADDR dest);
BFINST3(ArraySet_constindex,        BFADDR arraybase,   BFVALUE index,      BFADDR source);
BFINST3(ArrayGet_constindex,        BFADDR arraybase,   BFVALUE index,      BFADDR dest);

#endif // BRAINFUCKED_ASSEMBLER_H
